import numpy as np

import openmdao.api as om


class ComputeOweExercise(om.ExplicitComponent):
    """
    Computes the aircraft structural mass based on its MTOW and wing mass
    """

    def setup(self):

        # Defining the input(s)

        # Defining the output(s)

        pass

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        pass
