def compute_cd0(wing_area):
    """
    Computes the profile drag of the aircraft based on the wing area

    :param wing_area: Wing area, in m2

    :return cd0: profile drag, no unit
    """

    # Wet area of the aircraft without the wings
    wing_area_ref = 13.50

    # Profile drag coefficient of the aircraft without the wings
    cd0_other = 0.022

    # Constant linking the wing profile drag to its wet area, and by extension, its reference area
    c = 0.0004

    # Computation of the profile drag
    cd0 = cd0_other * wing_area_ref / wing_area + c

    return cd0
